/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2010 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* utility functions for eXtended FEM
 * $Id: xfem.c,v 1.38 2022/06/03 08:40:33 zlb Exp $ */

#include "phg.h"
#include "phg/xfem.h"

#include <math.h>
#include <strings.h>	/* bzero() */
#include <string.h>	/* memcpy() */

/* parameters with cmdline opptions */
typedef struct {
/* 
 * "ctol" is the tolerance on the local curvature: h * kapa <= ctol,
 * where kappa is the max absolute curvature of the interface in the element,
 * and h the diameter of the element. If ctol <= 0 then the curvatures are
 * not checked
 *
 * "dtol" is the interface deviation tolerance: interface deviations <= dtol.
 * If "dtol" <= 0.0 then interface deviations are not checked.
 *
 * "etol" is the edge fraction on a cut edge below which the element
 * is considered "small". If "etol" <= 0 then edge cuts are not checked.
 *
 * "vtol" is the volume fraction in a cut element below which the element
 * is considered "small". If "vtol" <= 0 then volumes are not checked.
 *
 * "horder" is the polynomial order used to approximate the Hessian of the
 * levelset function when estimating curvatures of the interface (only if the
 * levelset function is non-polynomial, otherwise it's unused). */
    FLOAT ctol, dtol, etol, vtol;
    INT refine_limit, horder;
    INT verb;		/* verbosity (-xfem_dbg_evel) */
    BOOLEAN vtk;
} XFEM_OPTIONS;

#ifndef PHG_TO_P4EST
XFEM_OPTIONS phg_xfem_options = {
	/* ctol */	0.5,
	/* dtol */	0.05,
	/* etol */	0.25,
	/* vtol */	0.0,
	/* refine_limit */ 20,
	/* horder */	1,
	/* verb */	0,
	/* vtk */	FALSE
};
#else	/* !defined(PHG_TO_P4EST) */
extern XFEM_OPTIONS phg_xfem_options;
#endif	/* !defined(PHG_TO_P4EST) */
#define Opts	phg_xfem_options

static FLOAT
get_curvature(XFEM_INFO *xi, ELEMENT *e, int np, const FLOAT pts[][Dim])
/* Returns approx(max(|k1|,|k2|)), k1 and k2 are the principal curvatures.
 * Note: the formulas for computing curvatures were taken from:
 *	Eric Albin, Ronnie Knikker, Shihe Xin, Christian Oliver Paschereit,
 *	Yves dAngelo.
 *
 *	Computational assessment of curvatures and principal
 *	directions of implicit surfaces from 3D scalar data.
 *
 *	Lecture Notes in Computer Science, Springer, 2017,
 *	Mathematical Methods for Curves and Surfaces, 10521, pp.1-22.
 *	10.1007/978-3-319-67885-6_1. hal-01486547
 * https://www.archives-ouvertes.fr/hal-01486547/document
 *
 * See get_deviation() for the arguments "np" and "pts".
 */
{
    int i, j, k;
#ifdef PHG_TO_P4EST
    FLOAT xref[] = {0.5, 0.5, 0.5};
#else
    FLOAT xref[] = {0.25, 0.25, 0.25, 0.25};
#endif
    FLOAT n[Dim], u[Dim], v[Dim], d, kh, kk, kmax;
    FLOAT hess[Dim][Dim], Luu, Luv, Lvv;

    if (np == 0) {
	pts = NULL;
	np = 1;
    }

    kmax = 0.0;
    for (i = 0; i < np; i++) {
	if (pts != NULL)
	    phgGeomXYZ2Lambda(xi->ls->g, e,
			pts[i][0], pts[i][1], pts[i][2], xref);
	phgDofEval(xi->ls_grad, e, xref, n);
	d = n[0] * n[0] + n[1] * n[1] + n[2] * n[2];
	/* The case below may occur when the surface is degenerated,
	 * or the barycenter is used (when pts == NULL) */
	if (d < FLOAT_EPSILON) {
#if 0
# warning
:wq
/* ./xfem_test -xfem_ctol 1000.0 -refine 4 -xfem_dbg_vtk -xfem_etol 0 */
	    phgInfo(-1, "*** %s: WARNING: interface possibly degenerated, "
			"element marked with -9999\n", __func__);
	    e->region_mark = -9999;
#else
	    static FLOAT d_min = FLOAT_EPSILON;
#if USE_OMP
# pragma omp threadprivate(d_min)
#endif	/* USE_OMP */
	    if (d < d_min) {
		/* note: pts may be NULL */
		FLOAT x[3];
		phgGeomLambda2XYZ(xi->ls->g, e, xref, x, x + 1, x + 2);
		phgInfo(-1, "*** %s: WARNING: the interface is possibly "
			    "degenerated at [%0.16g,%0.16g,%0.16g], "
			    "ignoring ctol for the element.\n",
			    __func__, x[0], x[1], x[2]);
		d_min = d;
	    }
#endif
	    return /*1e10*/0.0;
	}
	d = 1.0 / Sqrt(d);	/* d := 1/|\grad L| = 1/|Ln| */
	n[0] *= d;
	n[1] *= d;
	n[2] *= d;

	u[0] = n[0];
	u[1] = -n[2];
	u[2] = n[1];

	v[0] = n[1] * u[2] - n[2] * u[1];
	v[1] = n[2] * u[0] - n[0] * u[2];
	v[2] = n[0] * u[1] - n[1] * u[0];

	/* Note:
	 * 	L_n = (\grad L)^T n = (\grad L)^T\grad L/|\grad L| = |\grad L|
	 *
	 * 	L_uu = u^T \hess(L) u
	 * 	L_uv = u^T \hess(L) v
	 * 	L_vv = v^T \hess(L) v */
	phgDofEval(xi->ls_hess, e, xref, hess[0]);
	Luu = Luv = Lvv = 0.0;
	for (j = 0; j < Dim; j++) {
	    for (k = 0; k < Dim; k++) {
		Luu += u[j] * hess[j][k] * u[k];
		Luv += u[j] * hess[j][k] * v[k];
		Lvv += v[j] * hess[j][k] * v[k];
	    }
	}
	/* The mean cavature:
	 * kh := (Luu + Lvv) / (2 * |L_n|) = (Luu + Lvv) * d / 2 */
	kh = (Luu + Lvv) * d * 0.5;
	/* The Gaussian cavature:
	 * kk := (Luu * Lvv - Luv^2) / |L_n|^2 = (Luu * Lvv - Luv^2) * d^2 */
	kk = (Luu * Lvv - Luv * Luv) * d * d;
	d = kh * kh - kk;
	if (d < 0.0)
	    d = 0.0;
	if (kh > 0.0)
	    d = kh + Sqrt(d);
	else
	    d = -kh - Sqrt(d);
	if (kmax < d)
	    kmax = d;
	if (xi->ls_hess->poly_order == 0)
	    break;	/* constant Hessian ==> constant curvatures */
    }

    return kmax;
}


static struct {
    DOF *ls, *grad;
    ELEMENT *e;
} dummy_ctx;
#if USE_OMP
#pragma omp threadprivate(dummy_ctx)
#endif	/* USE_OMP */

static void
dummy_ls(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
    FLOAT xref[Dim + 1];

    if (!SpecialDofUserFunction(dummy_ctx.ls->userfunc)) {
	dummy_ctx.ls->userfunc(x, y, z, value);
	return;
    }
    phgGeomXYZ2Lambda(dummy_ctx.ls->g, dummy_ctx.e, x, y, z, xref);
    phgDofEval(dummy_ctx.ls, dummy_ctx.e, xref, value);
}

static void
dummy_grad(FLOAT x, FLOAT y, FLOAT z, FLOAT *values)
{
    FLOAT xref[Dim + 1];

    if (!SpecialDofUserFunction(dummy_ctx.grad->userfunc)) {
	dummy_ctx.grad->userfunc(x, y, z, values);
	return;
    }
    phgGeomXYZ2Lambda(dummy_ctx.grad->g, dummy_ctx.e, x, y, z, xref);
    phgDofEval(dummy_ctx.grad, dummy_ctx.e, xref, values);
}

static int horder = 0;

static void
resolve_interface(XFEM_INFO *xi, int quad_order)
/* refines the mesh such that constraints on the intersection types and
 * local curvature are satisfied.
 */
{
    GRID *g = xi->ls->g;
    ELEMENT *e;
    DOF *tmp;
    double t0;
    int i;

    if (Opts.ctol <= 0.0)
	return;			/* do nothing */

    horder = Opts.horder;
    assert(xi->ls_grad->poly_order >= 0 || horder >= 0);
    if (xi->ls_grad->poly_order >= 0)
	horder = xi->ls_grad->poly_order - 1;
    tmp = phgDofCopy(xi->ls_grad, NULL, DOF_DGn[horder + 1], "G");
    xi->ls_hess = phgDofGradient(tmp, NULL, NULL, "Hessian");
    phgDofFree(&tmp);
    xi->ls_hess->userfunc = DofInterpolation;
    if (Opts.verb > 0)
	phgPrintf("*** %s: ctol = %lg\n", __func__, (double)Opts.ctol);
    t0 = phgGetTime(NULL);
    while (TRUE) {
	double t = phgGetTime(NULL);
	INT n0, count = 0;	/* # of elements marked for refinement */
	ForAllElements(g, e) {
	    int np;		/* # of recorded intersection points */
	    FLOAT pts[2 * NEdge][Dim];
	    FLOAT corners[NVert][Dim];

	    dummy_ctx.ls = xi->ls;
	    dummy_ctx.grad = xi->ls_grad;
	    dummy_ctx.e = e;
	    phgGeomGetCorners(g, e, corners);

	    e->mark = 0;

	    /* First, a quick check on the element */
	    if (phgQuadInterfaceMarkElement(dummy_ls, xi->ls->poly_order,
			dummy_grad, e->elem_type, corners) != 0)
		continue;	/* no intersection is possible */

	    /* Check intersection with the edges. TODO: maintain a list of
	     * processed edges to save and reuse the cuts. */
	    np = 0;
	    for (i = 0; i < NEdge; i++) {
		int i0 = GetEdgeVertex(i, 0), i1 = GetEdgeVertex(i, 1);
#ifndef PHG_TO_P4EST
		FLOAT *v0 = corners[i0], *v1 = corners[i1];
#else	/* PHG_TO_P4EST */
		FLOAT v0[Dim] = {corners[(i0 >> 0) & 1][0],
				 corners[(i0 >> 1) & 1][1],
				 corners[(i0 >> 2) & 1][2]};
		FLOAT v1[Dim] = {corners[(i1 >> 0) & 1][0],
				 corners[(i1 >> 1) & 1][1],
				 corners[(i1 >> 2) & 1][2]};
#endif	/* PHG_TO_P4EST */
		FLOAT *cuts;
		int n;
		n = phgQuadInterfaceGetEdgeCuts(dummy_ls, xi->ls->poly_order,
						dummy_grad, v0, v1, 0.0, &cuts);
		if (n == 0) {
		    phgFree(cuts);
		    continue;
		}
		assert(np + n <= sizeof(pts) / sizeof(pts[0]));
		while (--n >= 0) {
		    pts[np][0] = v0[0] + cuts[n] * (v1[0] - v0[0]);
		    pts[np][1] = v0[1] + cuts[n] * (v1[1] - v0[1]);
		    pts[np][2] = v0[2] + cuts[n] * (v1[2] - v0[2]);
		    np++;
		}
		phgFree(cuts);
	    }

	    /* Check curvature */
	    if (phgGeomGetDiameter(g, e) * get_curvature(xi, e, np, pts)
								<= Opts.ctol)
		continue;
#if 0
#warning
FLOAT (*corners)[Dim] = phgGeomGetCorners(g, e, NULL);
phgInfo(-1, "**** element [[%g;%g;%g],[%g;%g;%g]], K = %g, h = %g\n", corners[0][0], corners[0][1], corners[0][2], corners[1][0], corners[1][1], corners[1][2], get_curvature(xi, e, np, pts), phgGeomGetDiameter(g, e));
#endif
	    e->mark = 1;
	    count += e->mark;
	}
#if USE_MPI
	if (g->nprocs > 1) {
	    INT count0 = count;
	    MPI_Allreduce(&count0, &count, 1, PHG_MPI_INT, MPI_SUM, g->comm);
	}
#endif	/* USE_MPI */
	if (Opts.verb > 0)
	    phgPrintf("*** %s: %"dFMT" elements to be refined, %0.2lgs.\n",
			__func__, count, phgGetTime(NULL) - t);
	if (count == 0)
	    break;
	n0 = g->nelem_global;
	phgRefineMarkedElements(g);
	phgBalanceGrid(g, 1.1, 0, NULL, 0.);
	if (n0 == g->nelem_global) {
	    /* marked elements not refined (p4est bug?). Test case:
	     * 		./interface-lv -xfem_ctol 4.0 -refine0 2 -verbosity 1
	     * (unit cube, LS = (4 * (x-xc)^2 + 2 * (y-yc)^2 + (z-zc)^2 - 1/50)^3 - (3 * (x-xc)^2 + 2 * (y-yc)^2) * (z-zc)^3 */
	    phgPrintf("*** %s: [FIXME] marked elements not refined, "
			"stop resolving interface.\n", __func__);
	    break;
	}
    }
    if (Opts.verb > 0)
	phgPrintf("*** %s: %"dFMT" elements, %0.2lgs.\n",
			__func__, g->nelem_global, phgGetTime(NULL) - t0);
    phgDofFree(&xi->ls_hess);
}

static void
mark_elements(XFEM_INFO *xi, int quad_order, DOF *old_info)
/* computes and stores element mark and volume fraction in xi->info[].mark
 * and xi->info[].frac.
 *
 * This function also initializes xi->info[*].macro[*] to -1, sets up
 * xi->work[*].{vmask,ecut}, and sets e->mark = 0.
 *
 * The argument "old_info", if not NULL, provides index, rank and generation
 * of the element or its parent in the mesh before refinement.
 */
{
    GRID *g = xi->ls->g;
    ELEMENT *e;
    XFEM_INFO *xi0;
    /* statistics on the minimum fractions of the subdomains */
    double tmp[2][phgMaxThreads], t0;
    INT copy_count = 0;
    INT etmp[3][phgMaxThreads];       /* ecnts: 0=- 1=0 2=+ */
    int ii;

    /* avoid recomputing if the mesh is unchanged */
    if (xi->marked && xi->serial_no == g->serial_no)
	return;

    phgSetupHalo(g, HALO_FACE);

    xi->serial_no = g->serial_no;
    xi->marked = TRUE;

    bzero(xi->minfrac0, sizeof(xi->minfrac0));
    bzero(xi->minfrac, sizeof(xi->minfrac));
    bzero(xi->ecnts, sizeof(xi->ecnts));

    /* allocate and initialize xi->info */
    if (xi->work == NULL && xi->info != NULL) {
	phgFree(xi->info);
	xi->info = NULL;
    }
    xi0 = phgCalloc(1, sizeof(*xi0));
    xi0->info = xi->info;
    xi0->work = xi->work;
#ifdef PHG_TO_P4EST
    xi->info = phgCalloc(g->nelem + g->nghost, sizeof(*xi->info));
    xi->work = phgCalloc(g->nelem + g->nghost, sizeof(*xi->work));
#else	/* defined(PHG_TO_P4EST) */
    xi->info = phgCalloc(g->nelem, sizeof(*xi->info));
    xi->work = phgCalloc(g->nelem, sizeof(*xi->work));
#endif	/* defined(PHG_TO_P4EST) */

    ForAllElements_(g, e, FOR_ALL) {
	int rank, generation;
	INT index;
	xi->work[e->index].frac = 2.0;	/* mark element as uninitialized */
#ifdef PHG_TO_P4EST
	xi->work[e->index].is_small[0] = xi->work[e->index].is_small[1] = 0;
#endif	/* defined(PHG_TO_P4EST) */
	if (xi0->info == NULL || old_info == NULL)
	    continue;
	/* copy xi->info[] entries for elements which have not been refined */
	index = (INT)(old_info->data[e->index * 2 + 1] + 0.5);
	rank = (int)(old_info->data[e->index * 2] + 0.5);
	generation = rank % 256;
	rank /= 256;
	if (rank != g->rank || generation != e->generation)
	    continue;
	memcpy(&xi->info[e->index], &xi0->info[index], sizeof(xi0->info[0]));
	memcpy(&xi->work[e->index], &xi0->work[index], sizeof(xi0->work[0]));
	assert(xi->work[e->index].frac >= 0 && xi->work[e->index].frac <=1.0);
	copy_count++;
    }
    phgFree(xi0->info);
    phgFree(xi0->work);
    phgFree(xi0);

    phgQuadInterfaceMarkGrid(xi->ls, xi->ls_grad);

    for (ii = 0; ii < phgMaxThreads; ii++) {
	tmp[0][ii] = tmp[1][ii] = 2.0;
	etmp[0][ii] = etmp[1][ii] = etmp[2][ii] = 0;
    }

    t0 = phgGetTime(NULL);
#if USE_OMP
#pragma omp parallel for private(e)
#endif  /* USE_OMP */
    ForAllElementsBegin_(g, e, FOR_ALL) {
	FLOAT *p, val;
	INT eno = e->index;
	int i, n;
	BOOLEAN initialized = (xi->work[eno].frac < 1.5);
#if defined(PHG_TO_P4EST)
	/* initalize "macro" members */
	xi->info[eno].macro[0] = xi->info[eno].macro[1] = -1;
#endif	/* defined(PHG_TO_P4EST) */

	/* set up "frac" and update "mark". */

	if (!initialized) {
	    FLOAT *rule_m = NULL, *rule_p = NULL;
	    if(e->mark == 0) {
		phgQuadInterface(xi->ls, xi->ls_grad, e, quad_order,
					&rule_m, NULL, &rule_p);
		if (phgQuadInterfaceRuleInfo(rule_p, NULL,NULL,NULL) == 0)
		    e->mark = -1;
		else if (phgQuadInterfaceRuleInfo(rule_m, NULL,NULL,NULL) == 0)
		    e->mark = 1;
	    }
	    xi->info[eno].mark = e->mark < 0 ? -1 : (e->mark > 0 ? 1 : 0);
	    if (xi->info[eno].mark == 0) {
		n = phgQuadInterfaceRuleInfo(rule_m, NULL, &p, NULL);
		for (i = 0, val = 0.; i < n; i++, p += Dim + 1)
		    val += p[Dim];
		xi->work[eno].frac = (val /= phgGeomGetVolume(g, e));
		if (xi->work[eno].frac < Sqrt(FLOAT_EPSILON))
		    xi->work[eno].frac = 0.0;
		else if (xi->work[eno].frac > 1.0 - Sqrt(FLOAT_EPSILON))
		    xi->work[eno].frac = 1.0;
	    }
	    else {
		xi->work[eno].frac = xi->info[eno].mark < 0 ? 1.0 : 0.0;
	    }
	    phgFree(rule_m);
	    phgFree(rule_p);
	}

	e->mark = 0;

	/* statistics about elements in different subdomains */
	if (xi->info[eno].mark < 0) {
	    if (tmp[0][phgThreadId] > 1.0)
		tmp[0][phgThreadId] = 1.0;
	    etmp[0][phgThreadId]++;
	}
	else if (xi->info[eno].mark > 0) {
	    if (tmp[1][phgThreadId] > 1.0)
		tmp[1][phgThreadId] = 1.0;
	    etmp[2][phgThreadId]++;
	}
	else {
	    val = xi->work[eno].frac;
	    if (val > 0. && tmp[0][phgThreadId] > val)
		tmp[0][phgThreadId] = val;
	    val = 1.0 - val;
	    if (val > 0. && tmp[1][phgThreadId] > val)
		tmp[1][phgThreadId] = val;
	    etmp[1][phgThreadId]++;
	}

    } ForAllElementsEnd

    xi->minfrac0[0] = xi->minfrac0[1] = 1.0;
    xi->ecnts[0] = xi->ecnts[1] = xi->ecnts[2] = 0;
    for (ii = 0; ii < phgMaxThreads; ii++) {
	if (xi->minfrac0[0] > tmp[0][ii])
	    xi->minfrac0[0] = tmp[0][ii];
	if (xi->minfrac0[1] > tmp[1][ii])
	    xi->minfrac0[1] = tmp[1][ii];
	xi->ecnts[0] += etmp[0][ii];
	xi->ecnts[1] += etmp[1][ii];
	xi->ecnts[2] += etmp[2][ii];
    }
#if USE_MPI
    if (g->nprocs > 1) {
	double tmp[2];
	INT etmp0[4], etmp1[4];
	tmp[0] = xi->minfrac0[0];
	tmp[1] = xi->minfrac0[1];
	MPI_Allreduce(tmp, xi->minfrac0, 2, MPI_DOUBLE, MPI_MIN, g->comm);
	etmp0[0] = xi->ecnts[0];
	etmp0[1] = xi->ecnts[1];
	etmp0[2] = xi->ecnts[2];
	etmp0[3] = copy_count;
	MPI_Allreduce(etmp0, etmp1, 4, PHG_MPI_INT, MPI_SUM, g->comm);
	xi->ecnts[0] = etmp1[0];
	xi->ecnts[1] = etmp1[1];
	xi->ecnts[2] = etmp1[2];
	copy_count = etmp1[3];
    }
#endif	/* USE_MPI */

    xi->minfrac[0] = xi->minfrac0[0];
    xi->minfrac[1] = xi->minfrac0[1];

    if (Opts.verb > 0)
	phgPrintf("*** %s: %"dFMT" new entries, %0.2lgs.\n", __func__,
			g->nelem_global - copy_count, phgGetTime(NULL) - t0);
}

#define Anchor(xi, g, e, sign) \
    (xi->info[e->index].macro[sign] < 0 ? e : \
		g->elems[xi->mlist[xi->info[e->index].macro[sign]][0]])


/*---------------------------------------------------------------------------*/

XFEM_INFO *
phgXFEMInit(DOF *ls, DOF *ls_grad, int ls_order, int quad_order)
/* This function creats and returns an XFEM_INFO object, and prepares the mesh
 * for XFEM computation by performing the following steps:
 *   1.	resolve the interface,
 *   2.	merge small slements.
 *
 * "ls"		is the levelset function
 * "ls_grad"	is the gradient of the levelset function (may be NULL)
 * "ls_order"	is the polynomial order of of the levelset function (only used
 *		if "ls"->type == DOF_ANALYTIC && "ls_order" > 0)
 * "quad_order"	is the numerical quadrature order for computing volume ratios.
 */
{
    XFEM_INFO *xi;
    DOF *old_info = NULL;
    double t0;
#ifdef PHG_TO_P4EST
#else	/* defined(PHG_TO_P4EST) */
    static BOOLEAN initialized = FALSE;

    if (!initialized) {
	if (ls != NULL)
	    phgError(1, "%s must be called with ls=NULL once before phgInit.\n",
		     __func__);
	initialized = TRUE;
	phgOptionsRegisterTitle("\nXFEM options:", "\n", "xfem");
	phgOptionsRegisterFloat("-xfem_ctol",
				"Tolerance on the curvature (h*kappa)",
				&Opts.ctol);
	phgOptionsRegisterFloat("-xfem_dtol",
				"Tolerance on the interface deviation",
				&Opts.dtol);
	phgOptionsRegisterFloat("-xfem_etol",
				"Edge fraction for small elements",
				&Opts.etol);
	phgOptionsRegisterFloat("-xfem_vtol",
				"Volume fraction for small elements",
				&Opts.vtol);
	phgOptionsRegisterInt("-xfem_horder",
				"Polynomial order for approximating Hessian of "
				"the levelset function", &Opts.horder);
	phgOptionsRegisterInt("-xfem_refine_limit",
				"Limit on the number of refinement passes when "
				"merging elements", &Opts.refine_limit);
	phgOptionsRegisterInt("-xfem_dbg_level",
				"Level of debugging messages", &Opts.verb);
	phgOptionsRegisterNoArg("-xfem_dbg_vtk",
				"Output VTK file for debugging", &Opts.vtk);
	return NULL;
    }
#endif	/* defined(PHG_TO_P4EST) */

    assert(ls != NULL);

    if (ls_order > 0) {
	if (ls->type == DOF_ANALYTIC)
	    phgDofSetPolyOrder(ls, ls_order);;
	if (ls_grad != NULL && ls_grad->type == DOF_ANALYTIC)
	    phgDofSetPolyOrder(ls_grad, ls_order - 1);;
    }

    xi = phgCalloc(1, sizeof(*xi));
    xi->ls = ls;
    if ((xi->ls_grad = ls_grad) == NULL) {
	xi->ls_grad = phgDofGradient(ls, NULL, NULL, NULL);
	xi->ls_grad->userfunc = DofInterpolation;
	xi->grad_owned = TRUE;
    }
    xi->serial_no = -1;

    resolve_interface(xi, quad_order);

    t0 = phgGetTime(NULL);

	mark_elements(xi, quad_order, old_info);

    if (Opts.verb > 0)
	phgPrintf("*** mark/merge elements: %0.2lgs\n", phgGetTime(NULL) - t0);

    return xi;
}

void
phgXFEMFree(XFEM_INFO **xi)
{

    if (xi == NULL || *xi == NULL)
	return;


    phgFree((*xi)->info);
    phgFree((*xi)->work);

    if ((*xi)->grad_owned)
	phgDofFree(&(*xi)->ls_grad);

    phgFree(*xi);
    *xi = NULL;
}


#define CheckGrid(xi) if ((xi)->serial_no != (xi)->ls->g->serial_no) \
    phgError(1, "%s:%d: grid changed after phgXFEMInit.\n", __FILE__, __LINE__);

QCACHE *
phgXFEMNewQC(XFEM_INFO *xi, void *fe, int subdomain_no)
/* Note: subdomain_no == 0: \Omega^-, subdomain_no == 1: \Omega^+ */
{
    QCACHE *qc = phgQCNew(QD_DEFAULT, fe);

    CheckGrid(xi)

    return qc;
}

#if defined(PHG_TO_P4EST)
INT
phgXFEMMapE2G(MAP *map, int dof_no, ELEMENT *e, int index)
{
    XFEM_INFO *xi = e->reserved;

    assert(xi == NULL || map->ndof % 2 == 0);

    if (xi != NULL) {
	CheckGrid(xi)
	/* DOFs must be ordered as (minus,plus) pairs */
	assert(map->dofs[2 * (dof_no / 2)]->type == 
	       map->dofs[2 * (dof_no / 2) + 1]->type);
	e = Anchor(xi, xi->ls->g, e, dof_no % 2);
    }

    return PHG_to_p8est_map_e2g(map, dof_no, e, index);
}
#endif	/* defined(PHG_TO_P4EST) */


void
phgXFEMSetInitialSolution(XFEM_INFO *xi, DOF *um, DOF *up)
{
    CheckGrid(xi)
}

void
phgXFEMUpdateSolution(XFEM_INFO *xi, DOF *um, DOF *up)
{
    CheckGrid(xi)
}


/*---------------------------------------------------------------------------*/

FLOAT
phgXFEMDot(XFEM_INFO *xi, DOF *um, DOF *up, DOF *vm, DOF *vp)
/* computes L2 dot-product (u,v), where
 *	u:=um and v:=vm in \Omega^-,
 * and
 *	u:=up and v:=vp in \Omega^+.
 */
{
    int dim = DofDim(um), Q_um, Q_vm, Q_up, Q_vp;
    FLOAT dot = 0.0;
    GRID *g = xi->ls->g;
    ELEMENT *e;
    QCACHE *qc_m, *qc_p;

    qc_m = phgQCNew(QD_DEFAULT, um);
    Q_um = phgQCAddFEFunction(qc_m, um);
    Q_vm = um == vm ? Q_um : phgQCAddFEFunction(qc_m, vm);

    qc_p = phgQCNew(QD_DEFAULT, up);
    Q_up = phgQCAddFEFunction(qc_p, up);
    Q_vp = up == vp ? Q_up : phgQCAddFEFunction(qc_p, vp);

    assert(dim == DofDim(up) && dim == DofDim(vm) && dim == DofDim(vp));

#if USE_OMP
#pragma omp parallel for private(e) reduction(+:dot)
#endif	/* USE_OMP */
    ForAllElementsBegin(g, e) {
	int i, j;
	FLOAT *rule_m = NULL, *rule_p = NULL;
	int order;	/* quadrature order */
	i = DofTypeOrder(um, e);
	if (i < DofTypeOrder(vm, e))
	    i = DofTypeOrder(vm, e);
	j = DofTypeOrder(up, e);
	if (j < DofTypeOrder(vp, e))
	    j = DofTypeOrder(vp, e);
	order = i + j + (xi->info[e->index].mark == 0 ? 3 : 0);
	phgQuadInterface(xi->info[e->index].mark == 0 ? xi->ls : NULL,
			xi->ls_grad, e, order,
			xi->info[e->index].mark <= 0 ? &rule_m : NULL, NULL,
			xi->info[e->index].mark >= 0 ? &rule_p : NULL);
	phgQCSetRule(qc_m, rule_m, -1.);
	dot += phgQCIntegrate(qc_m, e->index, Q_um, 0, qc_m, e->index, Q_vm, 0);
	phgFree(rule_m);

	phgQCSetRule(qc_p, rule_p, -1.);
	dot += phgQCIntegrate(qc_p, e->index, Q_up, 0, qc_p, e->index, Q_vp, 0);
	phgFree(rule_p);
    } ForAllElementsEnd

    phgQCFree(&qc_m);
    phgQCFree(&qc_p);


#if USE_MPI
    if (g->nprocs > 1) {
	FLOAT d = dot;
	MPI_Allreduce(&d, &dot, 1, PHG_MPI_FLOAT, PHG_SUM, g->comm);
    }
#endif	/* USE_MPI */

    return dot;
}

#ifndef PHG_TO_P4EST
void
phgXFEMProcessEmptyRows(SOLVER *solver)
/* process empty row */
{
    int i;
    MAT *A = phgSolverGetMat(solver);
    MAT_ROW *row;

    phgMatAssemble(A);
    phgMatDisassemble(A);
    assert(A->type == PHG_UNPACKED);
    for (i = 0; i < A->rmap->nlocal; i++) {
	row = A->rows + i;
	if (row->ncols == 0) {
	    /* empty row */
	    row->ncols = row->alloc = 1;
	    phgFree(row->cols);
	    row->cols = phgAlloc(sizeof(*row->cols));
	    row->cols[0] = i;
	    phgFree(row->data);
	    row->data = phgAlloc(sizeof(*row->data));
	    row->data[0] = 1.0;
	}
    }
    phgMatAssemble(A);
    phgMatDestroy(&A);
}
#endif	/* !defined(PHG_TO_P4EST) */
